#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <qdos.h>
#include <qptr.h>
#include <things.h>

#include "csrvthg.h"

char _prog_name[] = "Demo Server";
long _stack = 3000;

extern int InitWindow (struct QD_text *, short *);
extern void Update (int, char *, short);

typedef struct
{
    char *cmd;
    int (*func) (int, char *, short);
} CMDTABLE;

static int getdat (int ctx, char *cmd, short rlen)
{
    int t;
    QD_TEXT(26) txt;
    char *p;
    int o;
    short l;

    t = mt_rclck ();
    p = cn_date ((char *) &txt, t);
    if (*cmd == 't')
    {
	l = 8;
	o = 12;
    }
    else
    {
	o = 0;
	l = 20;
    }
    return SrvRply (ctx, p + o, l);
}

static int getnum (int ctx, char *cmd, short rlen)
{
    static unsigned long n = 0;
    char txt[16];

    itoa (n, txt);
    n++;
    return SrvRply (ctx, txt, strlen (txt));
}

extern char *_sys_var;

#define A 16807L
#define M 2147483647L		/* Mersenne prime 2^31 -1 */
#define Q 127773L		/* M div A (M / A) */
#define R 2836L			/* M mod A (M % A) */

static long pseudorand (void)
{
    long hi, lo;
    static long seed = 0;

    if (!seed || seed == M)
    {
	seed = mt_rclck() ^ _Jobid;
    }
    hi = seed / Q;
    lo = seed % Q;
    if ((seed = A * lo - R * hi) <= 0)
	seed += M;
    return seed;
}

static int getrnd (int ctx, char *cmd, short rlen)
{
    char txt[16];
    itoa (pseudorand (), txt);
    return SrvRply (ctx, txt, strlen (txt));
}
static int shuffle (int ctx, char *cmd, short rlen)
{
    char *p = cmd + 7;

    *p = *(cmd + rlen - 1);
    return SrvRply (ctx, p, rlen - 8);
}

static CMDTABLE ctbl[] =
{
    {"date", getdat},
    {"number", getnum},
    {"random", getrnd},
    {"shuffle", shuffle},
    {"time", getdat},
    {NULL}
};

#ifndef NICE

/*
 * It has been brought to my notice that on some systems, particularly those
 * where the OS name ends in a even digit, that a FreeSrvThg (SRVR,"demo");
 * call may not cause all clients to notice the demise of their server.
 *
 * So here's an unsupported way to use the undocumented CSM Usage list
 * structure to make sure all the little beggars get the message.
 *
 */

typedef struct __csm_use
{
    struct  __csm_use *next;
    long    OSprivate[3];
    long    flag;
    long    jobid;
    long    chi;
    long    cho;
    short   namlen;
    char    nam[1];
} CSM_USAGE;

static KillEmOff(void)
{
    static char csm[] = "Client Server";
    THING_LINKAGE *p;

    for( p = *((THING_LINKAGE **)(_sys_var + 0xb8)); p; p = p->th_nxtth)
    {
        if(p->th_name == (sizeof(csm) - 1))
        {
            if(strnicmp(csm, p->th_name_text, (sizeof(csm) - 1)) == 0)
            {
                CSM_USAGE *c, *k;
                for( c = (CSM_USAGE *)p->th_usage; c; c = k)
                {
                    k = c->next;
                    if(c->namlen == 4 && *(long *)c->nam == *(long *)"demo")
                    {
                        if(c->flag != SRVR)
                        {
                            mt_frjob(c->jobid, -10);
                        }
                    }
                }
            }
        }
    }
}
#endif
int main (int ac, char **av)
{
    static QD_TEXTI (Server, "Server");
    int ctx;
    char *sp;
    short nc;
    CMDTABLE *pct;
    int stdchn;
    short done = 0;
    char *p;

    if ((stdchn = InitWindow ((struct QD_text *) &Server, &done)) > 0)
    {
	if ((ctx = UseSrvThgL (SRVR, "demo", 256)) == 0)
	{
	    short rlen;
            short ncnt = 0;
            int nok = 0;
	    while (!done)
	    {
		ctx = SrvPoll ("demo", 5, &p, &rlen);
		if (ctx >= 0)
		{
		    nok++;
                    if(p && rlen)
                    {
		    if (io_sstrg (stdchn, 0, "C-SRVR: ", 8) > 0)
		    {
			Update (stdchn, p, rlen);
		    }

		    for (pct = ctbl; pct->cmd; pct++)
		    {
			if (strnicmp (pct->cmd, p, strlen (pct->cmd)) == 0)
			{
			    (pct->func) (ctx, p, rlen);
			    break;
			}
		    }
		    if (pct->cmd == NULL)
		    {
			SrvRply (ctx, "Error", 5);
		    }
		    free (p);
		    }
		}
		else if(ctx != -1)
		{
		    done = 1;
		}
                ncnt++;
                if(ncnt == 10)
                {
                    static int nmsg = 0;
                    char bmsg[80] = "CSRV Events : ";
                    char *p;
                    nmsg += ncnt;
                    itoa(nok, bmsg + 13);
                    strcat (bmsg, " / ");
                    p = bmsg + strlen(bmsg);
                    itoa(nmsg, p);
                    SrvBro("demo", bmsg, -1);
                    ncnt = 0;
        	}
	    }
            ctx = 0;
#ifndef NICE
            KillEmOff();
#endif
            FreeSrvThg (SRVR,"demo");
	}
    }
    return ctx;
}
int (*_Cstart) (int, char **) = main;

